# This file is part of pyotgw.
#
# pyotgw is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyotgw is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyotgw.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 Milan van Nugteren
#
"""pyotgw is a library to interface with the OpenTherm Gateway."""


import asyncio
import logging
import serial
import serial_asyncio
from asyncio import TimeoutError
from datetime import datetime

from . import protocol as otgw
from .vars import *

_LOGGER = logging.getLogger(__name__)


class pyotgw:

    def __init__(self):
        """Create a pyotgw object."""
        self._connected = False
        self._conn_error = False
        self._notify = []
        return

    async def connect(self, loop, port, baudrate=9600,
                      bytesize=serial.EIGHTBITS, parity=serial.PARITY_NONE,
                      stopbits=serial.STOPBITS_ONE, connection_timeout=10,
                      inactivity_timeout=5):
        """
        Connect to Opentherm Gateway at @port.
        Initialize the parameters obtained from the PS= and PR=
        commands and returns the status dict with the obtained values.
        If called while connected, reconnect to the gateway.

        This method is a coroutine
        """
        if self._connected:
            # We are actually reconnecting, cleanup first.
            _LOGGER.debug("Reconnecting to serial device on %s", port)
            if self._gpio_task:
                self._gpio_task.cancel()
            self._connected = False
        self.loop = loop
        try:
            transport, protocol = (
                await serial_asyncio.create_serial_connection(
                    loop, otgw.protocol, port, baudrate, bytesize, parity,
                    stopbits, connection_timeout))
        except serial.serialutil.SerialException as e:
            if not self._conn_error:
                _LOGGER.error(
                    "Could not connect to serial device on %s. "
                    "Will keep trying. Reported error was: %s", port, e)
                self._conn_error = True
            await asyncio.sleep(5)
            return await self.connect(loop, port, baudrate, bytesize, parity,
                                      stopbits, connection_timeout,
                                      inactivity_timeout)
        self._conn_error = False
        _LOGGER.debug("Connected to serial device on %s", port)
        self._transport = transport
        self._protocol = protocol
        self.loop.create_task(self._protocol.set_update_cb(self.send_report))
        if 0 < inactivity_timeout < 3:
            _LOGGER.error("Inactivity timeout too low. Should be at least 3 "
                          "seconds, got %d", inactivity_timeout)
        if inactivity_timeout >= 3:
            async def reconnect():
                """Reconnect to the OpenTherm Gateway."""
                _LOGGER.debug("Scheduling reconnect...")
                await self.connect(
                    loop, port, baudrate, bytesize, parity, stopbits,
                    connection_timeout, inactivity_timeout)
            self.loop.create_task(
                self._protocol.setup_watchdog(reconnect, inactivity_timeout))
        self._gpio_task = None
        self._connected = True
        await self.get_reports()
        await self.get_status()
        if (self._protocol.status.get(OTGW_GPIO_A)
                or self._protocol.status.get(OTGW_GPIO_B)):
            await self._poll_gpio(True)
        return dict(self._protocol.status)

    def get_room_temp(self):
        """
        Get the current room temperature.
        """
        if not self._connected:
            return
        return self._protocol.status.get(DATA_ROOM_TEMP)

    def get_target_temp(self):
        """
        Get the target temperature.
        """
        if not self._connected:
            return
        temp_ovrd = self._protocol.status.get(DATA_ROOM_SETPOINT_OVRD)
        if temp_ovrd:
            return temp_ovrd
        return self._protocol.status.get(DATA_ROOM_SETPOINT)

    async def set_target_temp(self, temp, temporary=True,
                              timeout=OTGW_DEFAULT_TIMEOUT):
        """
        Set the thermostat setpoint and return the newly accepted
        value.
        kwarg @temporary specifies whether or not the thermostat
            program may override this temperature.

        This method is a coroutine
        """
        cmd = (OTGW_CMD_TARGET_TEMP if temporary
               else OTGW_CMD_TARGET_TEMP_CONST)
        value = '{:2.1f}'.format(temp)
        status = {}
        ret = await self._wait_for_cmd(cmd, value, timeout)
        if ret is None:
            return
        ret = float(ret)
        if 0 <= ret <= 30:
            if ret == 0:
                status[OTGW_SETP_OVRD_MODE] = OTGW_SETP_OVRD_DISABLED
                status[DATA_ROOM_SETPOINT_OVRD] = None
            else:
                if temporary:
                    ovrd_mode = OTGW_SETP_OVRD_TEMPORARY
                else:
                    ovrd_mode = OTGW_SETP_OVRD_PERMANENT
                status[OTGW_SETP_OVRD_MODE] = ovrd_mode
                status[DATA_ROOM_SETPOINT_OVRD] = ret
            self._update_status(status)
            return ret

    def get_outside_temp(self):
        """
        Return the outside temperature as known in the gateway.
        """
        if not self._connected:
            return
        return self._protocol.status.get(DATA_OUTSIDE_TEMP)

    async def set_outside_temp(self, temp, timeout=OTGW_DEFAULT_TIMEOUT):
        """
        Configure the outside temperature to send to the thermostat.
        Allowed values are between -40.0 and +64.0, although
        thermostats may not display the full range. Specify a value
        above 64 (suggestion: 99) to clear a previously configured
        value.
        Return the accepted value on success, '-' if a previously
        configured value has been cleared or None on failure.

        This method is a coroutine
        """
        cmd = OTGW_CMD_OUTSIDE_TEMP
        status = {}
        if temp < -40:
            return None
        value = '{:2.1f}'.format(temp)
        ret = await self._wait_for_cmd(cmd, value, timeout)
        if ret is None:
            return
        if ret not in ['-', None]:
            ret = float(ret)
        if ret == '-':
            status[DATA_OUTSIDE_TEMP] = 0.0
        else:
            status[DATA_OUTSIDE_TEMP] = ret
        self._update_status(status)
        return ret

    async def set_clock(self, date=datetime.now(),
                        timeout=OTGW_DEFAULT_TIMEOUT):
        """
        Change the time and day of the week of the thermostat. The
        gateway will send the specified time and day of the week in
        response to the next time and date message from the thermostat.
        @date is a :datetime: object which defaults to now()
        Return the response from the gateway with format HH:MM/DOW,
        where DOW is a single digit: 1=Monday, 7=Sunday.

        This method is a coroutine
        """
        cmd = OTGW_CMD_SET_CLOCK
        value = "{}/{}".format(date.strftime('%H:%M'), date.isoweekday())
        return await self._wait_for_cmd(cmd, value, timeout)

    def get_hot_water_ovrd(self):
        """
        Return the current hot water override mode if set, otherwise
        None.
        """
        if not self._connected:
            return
        return self._protocol.status.get(OTGW_DHW_OVRD)

    async def get_reports(self):
        """
        Update the pyotgw object with the information from all of the
        PR commands and return the updated status dict.

        This method is a coroutine
        """
        cmd = OTGW_CMD_REPORT
        reports = {}
        for value in OTGW_REPORTS.keys():
            ret = await self._wait_for_cmd(cmd, value)
            if ret is None:
                reports[value] = None
                continue
            reports[value] = ret[2:]
        status = {
            OTGW_ABOUT: reports.get(OTGW_REPORT_ABOUT),
            OTGW_BUILD: reports.get(OTGW_REPORT_BUILDDATE),
            OTGW_CLOCKMHZ: reports.get(OTGW_REPORT_CLOCKMHZ),
            OTGW_MODE: reports.get(OTGW_REPORT_GW_MODE),
            OTGW_SMART_PWR: reports.get(OTGW_REPORT_SMART_PWR),
            OTGW_THRM_DETECT: reports.get(OTGW_REPORT_THERMOSTAT_DETECT),
            OTGW_DHW_OVRD: reports.get(OTGW_REPORT_DHW_SETTING),
        }
        ovrd_mode = reports.get(OTGW_REPORT_SETPOINT_OVRD)
        if ovrd_mode is not None:
            ovrd_mode = str.upper(ovrd_mode[0])
            status.update({OTGW_SETP_OVRD_MODE: ovrd_mode})
        gpio_funcs = reports.get(OTGW_REPORT_GPIO_FUNCS)
        if gpio_funcs is not None:
            status.update({
                OTGW_GPIO_A: int(gpio_funcs[0]),
                OTGW_GPIO_B: int(gpio_funcs[1]),
            })
        led_funcs = reports.get(OTGW_REPORT_LED_FUNCS)
        if led_funcs is not None:
            status.update({
                OTGW_LED_A: led_funcs[0],
                OTGW_LED_B: led_funcs[1],
                OTGW_LED_C: led_funcs[2],
                OTGW_LED_D: led_funcs[3],
                OTGW_LED_E: led_funcs[4],
                OTGW_LED_F: led_funcs[5],
            })
        tweaks = reports.get(OTGW_REPORT_TWEAKS)
        if tweaks is not None:
            status.update({
                OTGW_IGNORE_TRANSITIONS: int(tweaks[0]),
                OTGW_OVRD_HB: int(tweaks[1]),
            })
        sb_temp = reports.get(OTGW_REPORT_SETBACK_TEMP)
        if sb_temp is not None:
            status.update({OTGW_SB_TEMP: float(sb_temp)})
        vref = reports.get(OTGW_REPORT_VREF)
        if vref is not None:
            status.update({OTGW_VREF: int(vref)})
        if (ovrd_mode is not None and ovrd_mode != OTGW_SETP_OVRD_DISABLED):
            status[DATA_ROOM_SETPOINT_OVRD] = float(
                reports[OTGW_REPORT_SETPOINT_OVRD][1:])
        self._update_status(status)
        return dict(self._protocol.status)

    async def get_status(self):
        """
        Update the pyotgw object with the information from the PS
        command and return the updated status dict.

        This method is a coroutine
        """
        cmd = OTGW_CMD_SUMMARY
        ret = await self._wait_for_cmd(cmd, 1)
        # Return to 'reporting' mode
        asyncio.ensure_future(self._wait_for_cmd(cmd, 0), loop=self.loop)
        if ret is None:
            return
        fields = ret[1].split(',')
        device_status = fields[0].split('/')
        master_status = device_status[0]
        slave_status = device_status[1]
        remote_params = fields[2].split('/')
        capmodlimits = fields[4].split('/')
        dhw_setp_bounds = fields[13].split('/')
        ch_setp_bounds = fields[14].split('/')
        status = {
            DATA_MASTER_CH_ENABLED: int(master_status[7]),
            DATA_MASTER_DHW_ENABLED: int(master_status[6]),
            DATA_MASTER_COOLING_ENABLED: int(master_status[5]),
            DATA_MASTER_OTC_ENABLED: int(master_status[4]),
            DATA_MASTER_CH2_ENABLED: int(master_status[3]),
            DATA_SLAVE_FAULT_IND: int(slave_status[7]),
            DATA_SLAVE_CH_ACTIVE: int(slave_status[6]),
            DATA_SLAVE_DHW_ACTIVE: int(slave_status[5]),
            DATA_SLAVE_FLAME_ON: int(slave_status[4]),
            DATA_SLAVE_COOLING_ACTIVE: int(slave_status[3]),
            DATA_SLAVE_CH2_ACTIVE: int(slave_status[2]),
            DATA_SLAVE_DIAG_IND: int(slave_status[1]),
            DATA_CONTROL_SETPOINT: float(fields[1]),
            DATA_REMOTE_TRANSFER_DHW: int(remote_params[0][7]),
            DATA_REMOTE_TRANSFER_MAX_CH: int(remote_params[0][6]),
            DATA_REMOTE_RW_DHW: int(remote_params[1][7]),
            DATA_REMOTE_RW_MAX_CH: int(remote_params[1][6]),
            DATA_SLAVE_MAX_RELATIVE_MOD: float(fields[3]),
            DATA_SLAVE_MAX_CAPACITY: int(capmodlimits[0]),
            DATA_SLAVE_MIN_MOD_LEVEL: int(capmodlimits[1]),
            DATA_ROOM_SETPOINT: float(fields[5]),
            DATA_REL_MOD_LEVEL: float(fields[6]),
            DATA_CH_WATER_PRESS: float(fields[7]),
            DATA_ROOM_TEMP: float(fields[8]),
            DATA_CH_WATER_TEMP: float(fields[9]),
            DATA_DHW_TEMP: float(fields[10]),
            DATA_OUTSIDE_TEMP: float(fields[11]),
            DATA_RETURN_WATER_TEMP: float(fields[12]),
            DATA_SLAVE_DHW_MAX_SETP: int(dhw_setp_bounds[0]),
            DATA_SLAVE_DHW_MIN_SETP: int(dhw_setp_bounds[1]),
            DATA_SLAVE_CH_MAX_SETP: int(ch_setp_bounds[0]),
            DATA_SLAVE_CH_MIN_SETP: int(ch_setp_bounds[1]),
            DATA_DHW_SETPOINT: float(fields[15]),
            DATA_MAX_CH_SETPOINT: float(fields[16]),
            DATA_TOTAL_BURNER_STARTS: int(fields[17]),
            DATA_CH_PUMP_STARTS: int(fields[18]),
            DATA_DHW_PUMP_STARTS: int(fields[19]),
            DATA_DHW_BURNER_STARTS: int(fields[20]),
            DATA_TOTAL_BURNER_HOURS: int(fields[21]),
            DATA_CH_PUMP_HOURS: int(fields[22]),
            DATA_DHW_PUMP_HOURS: int(fields[23]),
            DATA_DHW_BURNER_HOURS: int(fields[24])
        }
        self._update_status(status)
        return dict(self._protocol.status)

    async def set_hot_water_ovrd(self, state, timeout=OTGW_DEFAULT_TIMEOUT):
        """
        Control the domestic hot water enable option. If the boiler has
        been configured to let the room unit control when to keep a
        small amount of water preheated, this command can influence
        that.
        @state should be 0 or 1 to enable the override in off or on
        state, or any other single character to disable the override.
        Return the accepted value, 'A' if the override is disabled
        or None on failure.

        This method is a coroutine
        """
        cmd = OTGW_CMD_HOT_WATER
        status = {}
        ret = await self._wait_for_cmd(cmd, state, timeout)
        if ret is 'A':
            status[OTGW_DHW_OVRD] = None
        elif ret in ['0', '1']:
            status[OTGW_DHW_OVRD] = int(ret)
        self._update_status(status)
        return ret

    def get_mode(self):
        """
        Return the last known gateway operating mode. Return "G" for
        Gateway mode or "M" for Monitor mode.
        """
        if not self._connected:
            return
        return self._protocol.status.get(OTGW_MODE)

    async def set_mode(self, mode, timeout=OTGW_DEFAULT_TIMEOUT):
        """
        Set the operating mode to either "Gateway" mode (:mode: =
        OTGW_MODE_GATEWAY or 1) or "Monitor" mode (:mode: =
        OTGW_MODE_MONITOR or 0), or use this method to reset the device
        (:mode: = OTGW_MODE_RESET).
        Return the newly activated mode, or the full renewed status
        dict after a reset.

        This method is a coroutine
        """
        cmd = OTGW_CMD_MODE
        status = {}
        ret = await self._wait_for_cmd(cmd, mode, timeout)
        if ret is None:
            return
        if mode is OTGW_MODE_RESET:
            self._protocol.status = {}
            await self.get_reports()
            await self.get_status()
            return dict(self._protocol.status)
        status[OTGW_MODE] = ret
        self._update_status(status)
        return ret

    def get_led_mode(self, led_id):
        """
        Return the led mode for led :led_id:.
        @led_id Character in range A-F
        """
        if not self._connected:
            return
        return self._protocol.status.get("OTGW_LED_{}".format(led_id))

    async def set_led_mode(self, led_id, mode, timeout=OTGW_DEFAULT_TIMEOUT):
        """
        Configure the functions of the six LEDs (A-F) that can
        optionally be connected to pins RB3/RB4/RB6/RB7 and the GPIO
        pins of the PIC. The following functions are currently
        available:

        R Receiving an Opentherm message from the thermostat or boiler
        X Transmitting an Opentherm message to the thermostat or boiler
        T Transmitting or receiving a message on the master interface
        B Transmitting or receiving a message on the slave interface
        O Remote setpoint override is active
        F Flame is on
        H Central heating is on
        W Hot water is on
        C Comfort mode (Domestic Hot Water Enable) is on
        E Transmission error has been detected
        M Boiler requires maintenance
        P Raised power mode active on thermostat interface.

        Return the new mode for the specified led, or None on failure.

        This method is a coroutine
        """
        if led_id in "ABCDEF" and mode in "RXTBOFHWCEMP":
            cmd = globals().get("OTGW_CMD_LED_{}".format(led_id))
            status = {}
            ret = await self._wait_for_cmd(cmd, mode, timeout)
            if ret is None:
                return
            var = globals().get("OTGW_LED_{}".format(led_id))
            status[var] = ret
            self._update_status(status)
            return ret

    def get_gpio_mode(self, gpio_id):
        """
        Return the gpio mode for gpio :gpio_id:.
        @gpio_id Character A or B.
        """
        if not self._connected:
            return
        return self._protocol.status.get("OTGW_GPIO_{}".format(gpio_id))

    async def set_gpio_mode(self, gpio_id, mode, timeout=OTGW_DEFAULT_TIMEOUT):
        """
        Configure the functions of the two GPIO pins of the gateway.
        The following functions are available:

        0 No function, default for both ports on a freshly flashed chip
        1 Ground - A permanently low output (0V). Could be used for a
            power LED
        2 Vcc - A permanently high output (5V). Can be used as a
            short-proof power supply for some external circuitry used
            by the other GPIO port
        3 LED E - An additional LED if you want to present more than 4
            LED functions
        4 LED F - An additional LED if you want to present more than 5
            LED functions
        5 Home - Set thermostat to setback temperature when pulled low
        6 Away - Set thermostat to setback temperature when pulled high
        7 DS1820 (GPIO port B only) - Data line for a DS18S20 or
            DS18B20 temperature sensor used to measure the outside
            temperature. A 4k7 resistor should be connected between
            GPIO port B and Vcc

        Return the new mode for the specified gpio, or None on
        failure.

        This method is a coroutine
        """
        if gpio_id in "AB" and mode in range(8):
            if mode == 7 and gpio_id != "B":
                return None
            cmd = globals().get("OTGW_CMD_GPIO_{}".format(gpio_id))
            status = {}
            ret = await self._wait_for_cmd(cmd, mode, timeout)
            if ret is None:
                return
            ret = int(ret)
            var = globals().get("OTGW_GPIO_{}".format(gpio_id))
            status[var] = ret
            self._update_status(status)
            asyncio.ensure_future(
                self._poll_gpio(self._protocol.status.get(OTGW_GPIO_A)
                                or self._protocol.status.get(OTGW_GPIO_B)))
            return ret

    def get_setback_temp(self):
        """
        Return the last known setback temperature from the device.
        """
        if not self._connected:
            return
        return self._protocol.status.get(OTGW_SB_TEMP)

    async def set_setback_temp(self, sb_temp, timeout=OTGW_DEFAULT_TIMEOUT):
        """
        Configure the setback temperature to use in combination with
        GPIO functions HOME (5) and AWAY (6).
        Return the new setback temperature, or None on failure.

        This method is a coroutine
        """
        cmd = OTGW_CMD_SETBACK
        status = {}
        ret = await self._wait_for_cmd(cmd, sb_temp, timeout)
        if ret is None:
            return
        ret = float(ret)
        status[OTGW_SB_TEMP] = ret
        self._update_status(status)
        return ret

    async def add_alternative(self, alt, timeout=OTGW_DEFAULT_TIMEOUT):
        """
        Add the specified Data-ID to the list of alternative commands
        to send to the boiler instead of a Data-ID that is known to be
        unsupported by the boiler. Alternative Data-IDs will always be
        sent to the boiler in a Read-Data request message with the
        data-value set to zero. The table of alternative Data-IDs is
        stored in non-volatile memory so it will persist even if the
        gateway has been powered off. Data-ID values from 1 to 255 are
        allowed.
        Return the ID that was added to the list, or None on failure.

        This method is a coroutine
        """
        cmd = OTGW_CMD_ADD_ALT
        alt = int(alt)
        if alt < 1 or alt > 255:
            return None
        ret = await self._wait_for_cmd(cmd, alt, timeout)
        if ret is not None:
            return int(ret)

    async def del_alternative(self, alt, timeout=OTGW_DEFAULT_TIMEOUT):
        """
        Remove the specified Data-ID from the list of alternative
        commands. Only one occurrence is deleted. If the Data-ID
        appears multiple times in the list of alternative commands,
        this command must be repeated to delete all occurrences. The
        table of alternative Data-IDs is stored in non-volatile memory
        so it will persist even if the gateway has been powered off.
        Data-ID values from 1 to 255 are allowed.
        Return the ID that was removed from the list, or None on
        failure.

        This method is a coroutine
        """
        cmd = OTGW_CMD_DEL_ALT
        alt = int(alt)
        if alt < 1 or alt > 255:
            return None
        ret = await self._wait_for_cmd(cmd, alt, timeout)
        if ret is not None:
            return int(ret)

    async def add_unknown_id(self, unknown_id, timeout=OTGW_DEFAULT_TIMEOUT):
        """
        Inform the gateway that the boiler doesn't support the
        specified Data-ID, even if the boiler doesn't indicate that
        by returning an Unknown-DataId response. Using this command
        allows the gateway to send an alternative Data-ID to the boiler
        instead.
        Return the added ID, or None on failure.

        This method is a coroutine
        """
        cmd = OTGW_CMD_UNKNOWN_ID
        unknown_id = int(unknown_id)
        if unknown_id < 1 or unknown_id > 255:
            return None
        ret = await self._wait_for_cmd(cmd, unknown_id, timeout)
        if ret is not None:
            return int(ret)

    async def del_unknown_id(self, unknown_id, timeout=OTGW_DEFAULT_TIMEOUT):
        """
        Start forwarding the specified Data-ID to the boiler again.
        This command resets the counter used to determine if the
        specified Data-ID is supported by the boiler.
        Return the ID that was removed, or None on failure.

        This method is a coroutine
        """
        cmd = OTGW_CMD_KNOWN_ID
        unknown_id = int(unknown_id)
        if unknown_id < 1 or unknown_id > 255:
            return None
        ret = await self._wait_for_cmd(cmd, unknown_id, timeout)
        if ret is not None:
            return int(ret)

    async def prio_message(self, data_id, timeout=OTGW_DEFAULT_TIMEOUT):
        """
        NOT IMPLEMENTED YET!
        Specify a one-time priority message to be sent to the boiler at
        the first opportunity. If the specified message returns the
        number of Transparent Slave Parameters (TSPs) or Fault History
        Buffers (FHBs), the gateway will proceed to request those TSPs
        or FHBs.

        This method is a coroutine
        """
        # TODO: implement this, including FHB/TSP processing
        return

    async def set_response(self, data_id, data, timeout=OTGW_DEFAULT_TIMEOUT):
        """
        NOT IMPLEMENTED YET!
        Configure a response to send back to the thermostat instead of
        the response produced by the boiler.
        @data is a list of either one or two hex byte values
        Return the data ID for which the response was set, or None on
        failure.

        This method is a coroutine
        """
        # TODO: implement this
        return

    async def clear_response(self, data_id, timeout=OTGW_DEFAULT_TIMEOUT):
        """
        Clear a previously configured response to send back to the
        thermostat for :data_id:.
        Return the data ID for which the response was cleared, or None
        on failure.

        This method is a coroutine
        """
        # TODO: implement this
        return

    async def set_max_ch_setpoint(self, temperature,
                                  timeout=OTGW_DEFAULT_TIMEOUT):
        """
        Set the maximum central heating setpoint. This command is only
        available with boilers that support this function.
        Return the newly accepted setpoint, or None on failure.

        This method is a coroutine
        """
        cmd = OTGW_CMD_SET_MAX
        status = {}
        ret = await self._wait_for_cmd(cmd, temperature, timeout)
        if ret is None:
            return
        ret = float(ret)
        status[DATA_MAX_CH_SETPOINT] = ret
        self._update_status(status)
        return ret

    async def set_dhw_setpoint(self, temperature,
                               timeout=OTGW_DEFAULT_TIMEOUT):
        """
        Set the domestic hot water setpoint. This command is only
        available with boilers that support this function.
        Return the newly accepted setpoint, or None on failure.

        This method is a coroutine
        """
        cmd = OTGW_CMD_SET_WATER
        status = {}
        ret = await self._wait_for_cmd(cmd, temperature, timeout)
        if ret is None:
            return
        ret = float(ret)
        status[DATA_DHW_SETPOINT] = ret
        self._update_status(status)
        return ret

    async def set_max_relative_mod(self, max_mod,
                                   timeout=OTGW_DEFAULT_TIMEOUT):
        """
        Override the maximum relative modulation from the thermostat.
        Valid values are 0 through 100. Clear the setting by specifying
        a non-numeric value.
        Return the newly accepted value, '-' if a previous value was
        cleared, or None on failure.

        This method is a coroutine
        """
        if isinstance(max_mod, int) and not 0 <= max_mod <= 100:
            return None
        cmd = OTGW_CMD_MAX_MOD
        status = {}
        ret = await self._wait_for_cmd(cmd, max_mod, timeout)
        if ret not in ['-', None]:
            ret = int(ret)
        if ret == '-':
            status[DATA_SLAVE_MAX_RELATIVE_MOD] = None
        else:
            status[DATA_SLAVE_MAX_RELATIVE_MOD] = ret
        self._update_status(status)
        return ret

    async def set_control_setpoint(self, setpoint,
                                   timeout=OTGW_DEFAULT_TIMEOUT):
        """
        Manipulate the control setpoint being sent to the boiler. Set
        to 0 to pass along the value specified by the thermostat.
        Return the newly accepted value, or None on failure.

        This method is a coroutine
        """
        cmd = OTGW_CMD_CONTROL_SETPOINT
        status = {}
        ret = await self._wait_for_cmd(cmd, setpoint, timeout)
        if ret is None:
            return
        ret = float(ret)
        status[DATA_CONTROL_SETPOINT] = ret
        self._update_status(status)
        return ret

    async def set_ch_enable_bit(self, ch_bit, timeout=OTGW_DEFAULT_TIMEOUT):
        """
        Control the CH enable status bit when overriding the control
        setpoint. By default the CH enable bit is set after a call to
        set_control_setpoint with a value other than 0. With this
        method, the bit can be manipulated.
        @ch_bit can be either 0 or 1.
        Return the newly accepted value (0 or 1), or None on failure.

        This method is a coroutine
        """
        if ch_bit not in [0, 1]:
            return None
        cmd = OTGW_CMD_CONTROL_HEATING
        status = {}
        ret = await self._wait_for_cmd(cmd, ch_bit, timeout)
        if ret is None:
            return
        ret = int(ret)
        status[DATA_MASTER_CH_ENABLED] = ret
        self._update_status(status)
        return ret

    async def set_ventilation(self, pct, timeout=OTGW_DEFAULT_TIMEOUT):
        """
        Configure a ventilation setpoint override value (0-100%).
        Return the newly accepted value, or None on failure.
        @pct :int: Must be between 0 and 100.

        This method is a coroutine
        """
        if not 0 <= pct <= 100:
            return None
        cmd = OTGW_CMD_VENT
        status = {}
        ret = await self._wait_for_cmd(cmd, pct, timeout)
        if ret is None:
            return
        ret = int(ret)
        status[DATA_COOLING_CONTROL] = ret
        self._update_status(status)
        return ret

    def subscribe(self, coro):
        """
        Subscribe to status updates from the Opentherm Gateway.
        Can only be used after connect()
        @coro is a coroutine which will be called with a single
        argument (status) when a status change occurs.
        Return True on success, False if not connected or already
        subscribed.
        """
        if coro not in self._notify:
            self._notify.append(coro)
            return True
        return False

    def unsubscribe(self, coro):
        """
        Unsubscribe from status updates from the Opentherm Gateway.
        Can only be used after connect()
        @coro is a coroutine which has been subscribed with subscribe()
        earlier.
        Return True on success, false if not connected or subscribed.
        """
        if coro in self._notify:
            self._notify.remove(coro)
            return True
        return False

    async def send_report(self, status):
        """
        Call all subscribed coroutines in _notify whenever a status
        update occurs.

        This method is a coroutine
        """
        if len(self._notify) > 0:
            # Each client gets its own copy of the dict.
            asyncio.gather(*[coro(dict(status)) for coro in self._notify],
                           loop=self.loop)

    async def _wait_for_cmd(self, cmd, value, timeout=OTGW_DEFAULT_TIMEOUT):
        """
        Wrap @cmd in applicable asyncio call.

        This method is a coroutine.
        """
        if not self._connected:
            return
        try:
            return await asyncio.wait_for(self._protocol.issue_cmd(cmd, value),
                                          timeout,
                                          loop=self.loop)
        except TimeoutError:
            _LOGGER.error("Timed out waiting for command: %s, value: %s.", cmd,
                          value)
            return
        except (RuntimeError, SyntaxError, ValueError) as exc:
            _LOGGER.error("Command %s with value %s raised exception: %s", cmd,
                          value, exc)

    async def _poll_gpio(self, poll, interval=10):
        """
        Start or stop polling GPIO states.

        GPIO states aren't being pushed by the gateway, we need to poll
        if we want updates.
        """
        if poll and self._gpio_task is None:
            async def polling_routine(interval):
                """Poll GPIO state every @interval seconds."""
                while True:
                    try:
                        pios = None
                        ret = await self._wait_for_cmd(
                            OTGW_CMD_REPORT, OTGW_REPORT_GPIO_STATES)
                        if ret:
                            pios = ret[2:]
                            status = {
                                OTGW_GPIO_A_STATE: int(pios[0]),
                                OTGW_GPIO_B_STATE: int(pios[1]),
                            }
                            self._update_status(status)
                        await asyncio.sleep(interval)
                    except asyncio.CancelledError:
                        status = {
                            OTGW_GPIO_A_STATE: 0,
                            OTGW_GPIO_B_STATE: 0,
                        }
                        self._update_status(status)
                        self._gpio_task = None
                        break
            self._gpio_task = self.loop.create_task(polling_routine(interval))
        elif not poll and self._gpio_task is not None:
            self._gpio_task.cancel()

    def _update_status(self, update):
        """Update the status dict and push it to subscribers."""
        if isinstance(update, dict):
            self._protocol.status.update(update)
            self._protocol._updateq.put_nowait(self._protocol.status)



MSG_STATUS = b'\x00'
MSG_TSET = b'\x01'
MSG_MCONFIG = b'\x02'
MSG_SCONFIG = b'\x03'
MSG_COMMAND = b'\x04'
MSG_ASFFLAGS = b'\x05'
MSG_RBPFLAGS = b'\x06'
MSG_COOLING = b'\x07'
MSG_TSETC2 = b'\x08'
MSG_TROVRD = b'\x09'
MSG_TSP = b'\x0A'
MSG_TSPIDX = b'\x0B'
MSG_FHBSIZE = b'\x0C'
MSG_FHBIDX = b'\x0D'
MSG_MAXRMOD = b'\x0E'
MSG_MAXCAPMINMOD = b'\x0F'
MSG_TRSET = b'\x10'
MSG_RELMOD = b'\x11'
MSG_CHPRESS = b'\x12'
MSG_DHWFLOW = b'\x13'
MSG_TIME = b'\x14'
MSG_DATE = b'\x15'
MSG_YEAR = b'\x16'
MSG_TRSET2 = b'\x17'
MSG_TROOM = b'\x18'
MSG_TBOILER = b'\x19'
MSG_TDHW = b'\x1A'
MSG_TOUTSIDE = b'\x1B'
MSG_TRET = b'\x1C'
MSG_TSTOR = b'\x1D'
MSG_TCOLL = b'\x1E'
MSG_TFLOWCH2 = b'\x1F'
MSG_TDHW2 = b'\x20'
MSG_TEXHAUST = b'\x21'
MSG_TDHWSETUL = b'\x30'
MSG_TCHSETUL = b'\x31'
MSG_OTCCURVEUL = b'\x32'
MSG_TDHWSET = b'\x38'
MSG_MAXTSET = b'\x39'
MSG_OTCCURVE = b'\x3A'
MSG_ROVRD = b'\x64'
MSG_OEMDIAG = b'\x73'
MSG_BURNSTARTS = b'\x74'
MSG_CHPUMPSTARTS = b'\x75'
MSG_DHWPUMPSTARTS = b'\x76'
MSG_DHWBURNSTARTS = b'\x77'
MSG_BURNHRS = b'\x78'
MSG_CHPUMPHRS = b'\x79'
MSG_DHWPUMPHRS = b'\x7A'
MSG_DHWBURNHRS = b'\x7B'
MSG_OTVERM = b'\x7C'
MSG_OTVERS = b'\x7D'
MSG_MVER = b'\x7E'
MSG_SVER = b'\x7F'


# MSG_STATUS
DATA_MASTER_CH_ENABLED = 'master_ch_enabled'
DATA_MASTER_DHW_ENABLED = 'master_dhw_enabled'
DATA_MASTER_COOLING_ENABLED = 'master_cooling_enabled'
DATA_MASTER_OTC_ENABLED = 'master_otc_enabled'
DATA_MASTER_CH2_ENABLED = 'master_ch2_enabled'
DATA_SLAVE_FAULT_IND = 'slave_fault_indication'
DATA_SLAVE_CH_ACTIVE = 'slave_ch_active'
DATA_SLAVE_DHW_ACTIVE = 'slave_dhw_active'
DATA_SLAVE_FLAME_ON = 'slave_flame_on'
DATA_SLAVE_COOLING_ACTIVE = 'slave_cooling_active'
DATA_SLAVE_CH2_ACTIVE = 'slave_ch2_active'
DATA_SLAVE_DIAG_IND = 'slave_diagnostic_indication'

# MSG_TSET
DATA_CONTROL_SETPOINT = 'control_setpoint'

# MSG_MCONFIG
DATA_MASTER_MEMBERID = 'master_memberid'

# MSG_SCONFIG
DATA_SLAVE_DHW_PRESENT = 'slave_dhw_present'
DATA_SLAVE_CONTROL_TYPE = 'slave_control_type'
DATA_SLAVE_COOLING_SUPPORTED = 'slave_cooling_supported'
DATA_SLAVE_DHW_CONFIG = 'slave_dhw_config'
DATA_SLAVE_MASTER_LOW_OFF_PUMP = 'slave_master_low_off_pump'
DATA_SLAVE_CH2_PRESENT = 'slave_ch2_present'
DATA_SLAVE_MEMBERID = 'slave_memberid'

# MSG_COMMAND

# MSG_ASFFLAGS
DATA_SLAVE_SERVICE_REQ = 'slave_service_required'
DATA_SLAVE_REMOTE_RESET = 'slave_remote_reset'
DATA_SLAVE_LOW_WATER_PRESS = 'slave_low_water_pressure'
DATA_SLAVE_GAS_FAULT = 'slave_gas_fault'
DATA_SLAVE_AIR_PRESS_FAULT = 'slave_air_pressure_fault'
DATA_SLAVE_WATER_OVERTEMP = 'slave_water_overtemp'
DATA_SLAVE_OEM_FAULT = 'slave_oem_fault'

# MSG_RBPFLAGS
DATA_REMOTE_TRANSFER_DHW = 'remote_transfer_dhw'
DATA_REMOTE_TRANSFER_MAX_CH = 'remote_transfer_max_ch'
DATA_REMOTE_RW_DHW = 'remote_rw_dhw'
DATA_REMOTE_RW_MAX_CH = 'remote_rw_max_ch'

# MSG_COOLING
DATA_COOLING_CONTROL = 'cooling_control'

# MSG_TSETC2
DATA_CONTROL_SETPOINT_2 = 'control_setpoint_2'

# MSG_TROVRD
DATA_ROOM_SETPOINT_OVRD = 'room_setpoint_ovrd'

# MSG_TSP

# MSG_TSPIDX

# MSG_FHBSIZE

# MSG_FHBIDX

# MSG_MAXRMOD
DATA_SLAVE_MAX_RELATIVE_MOD = 'slave_max_relative_modulation'

# MSG_MAXCAPMINMOD
DATA_SLAVE_MAX_CAPACITY = 'slave_max_capacity'
DATA_SLAVE_MIN_MOD_LEVEL = 'slave_min_mod_level'

# MSG_TRSET
DATA_ROOM_SETPOINT = 'room_setpoint'

# MSG_RELMOD
DATA_REL_MOD_LEVEL = 'relative_mod_level'

# MSG_CHPRESS
DATA_CH_WATER_PRESS = 'ch_water_pressure'

# MSG_DHWFLOW
DATA_DHW_FLOW_RATE = 'dhw_flow_rate'

# MSG_TIME

# MSG_DATE

# MSG_YEAR

# MSG_TRSET2
DATA_ROOM_SETPOINT_2 = 'room_setpoint_2'

# MSG_TROOM
DATA_ROOM_TEMP = 'room_temp'

# MSG_TBOILER
DATA_CH_WATER_TEMP = 'ch_water_temp'

# MSG_TDHW
DATA_DHW_TEMP = 'dhw_temp'

# MSG_TOUTSIDE
DATA_OUTSIDE_TEMP = 'outside_temp'

# MSG_TRET
DATA_RETURN_WATER_TEMP = 'return_water_temp'

# MSG_TSTOR
DATA_SOLAR_STORAGE_TEMP = 'solar_storage_temp'

# MSG_TCOLL
DATA_SOLAR_COLL_TEMP = 'solar_coll_temp'

# MSG_TFLOWCH2
DATA_CH_WATER_TEMP_2 = 'ch_water_temp_2'

# MSG_TDHW2
DATA_DHW_TEMP_2 = 'dhw_temp_2'

# MSG_TEXHAUST
DATA_EXHAUST_TEMP = 'exhaust_temp'

# MSG_TDHWSETUL
DATA_SLAVE_DHW_MAX_SETP = 'slave_dhw_max_setp'
DATA_SLAVE_DHW_MIN_SETP = 'slave_dhw_min_setp'

# MSG_TCHSETUL
DATA_SLAVE_CH_MAX_SETP = 'slave_ch_max_setp'
DATA_SLAVE_CH_MIN_SETP = 'slave_ch_min_setp'

# MSG_OTCCURVEUL

# MSG_TDHWSET
DATA_DHW_SETPOINT = 'dhw_setpoint'

# MSG_MAXTSET
DATA_MAX_CH_SETPOINT = 'max_ch_setpoint'

# MSG_OTCCURVE

# MSG_ROVRD
DATA_ROVRD_MAN_PRIO = 'rovrd_man_prio'
DATA_ROVRD_AUTO_PRIO = 'rovrd_auto_prio'

# MSG_OEMDIAG
DATA_OEM_DIAG = 'oem_diag'

# MSG_BURNSTARTS
DATA_TOTAL_BURNER_STARTS = 'burner_starts'

# MSG_CHPUMPSTARTS
DATA_CH_PUMP_STARTS = 'ch_pump_starts'

# MSG_DHWPUMPSTARTS
DATA_DHW_PUMP_STARTS = 'dhw_pump_starts'

# MSG_DHWBURNSTARTS
DATA_DHW_BURNER_STARTS = 'dhw_burner_starts'

# MSG_BURNHRS
DATA_TOTAL_BURNER_HOURS = 'burner_hours'

# MSG_CHPUMPHRS
DATA_CH_PUMP_HOURS = 'ch_pump_hours'

# MSG_DHWPUMPHRS
DATA_DHW_PUMP_HOURS = 'dhw_pump_hours'

# MSG_DHWBURNHRS
DATA_DHW_BURNER_HOURS = 'dhw_burner_hours'

# MSG_OTVERM
DATA_MASTER_OT_VERSION = 'master_ot_version'

# MSG_OTVERS
DATA_SLAVE_OT_VERSION = 'slave_ot_version'

# MSG_MVER
DATA_MASTER_PRODUCT_TYPE = 'master_product_type'
DATA_MASTER_PRODUCT_VERSION = 'master_product_version'

# MSG_SVER
DATA_SLAVE_PRODUCT_TYPE = 'slave_product_type'
DATA_SLAVE_PRODUCT_VERSION = 'slave_product_version'


READ_DATA = 0x0
WRITE_DATA = 0x1
INVALID_DATA = 0x2
RESERVED = 0x3
READ_ACK = 0x4
WRITE_ACK = 0x5
DATA_INVALID = 0x6
UNKNOWN_DATAID = 0x7

OTGW_DEFAULT_TIMEOUT = 3

OTGW_CMD_TARGET_TEMP = 'TT'
OTGW_CMD_TARGET_TEMP_CONST = 'TC'
OTGW_CMD_OUTSIDE_TEMP = 'OT'
OTGW_CMD_SET_CLOCK = 'SC'
OTGW_CMD_HOT_WATER = 'HW'
OTGW_CMD_REPORT = 'PR'
OTGW_CMD_SUMMARY = 'PS'
OTGW_CMD_MODE = 'GW'
OTGW_CMD_LED_A = 'LA'
OTGW_CMD_LED_B = 'LB'
OTGW_CMD_LED_C = 'LC'
OTGW_CMD_LED_D = 'LD'
OTGW_CMD_LED_E = 'LE'
OTGW_CMD_LED_F = 'LF'
OTGW_CMD_GPIO_A = 'GA'
OTGW_CMD_GPIO_B = 'GB'
OTGW_CMD_SETBACK = 'SB'
OTGW_CMD_ADD_ALT = 'AA'
OTGW_CMD_DEL_ALT = 'DA'
OTGW_CMD_UNKNOWN_ID = 'UI'
OTGW_CMD_KNOWN_ID = 'KI'
OTGW_CMD_PRIO_MSG = 'PM'
OTGW_CMD_SET_RESP = 'SR'
OTGW_CMD_CLR_RESP = 'CR'
OTGW_CMD_SET_MAX = 'SH'
OTGW_CMD_SET_WATER = 'SW'
OTGW_CMD_MAX_MOD = 'MM'
OTGW_CMD_CONTROL_SETPOINT = 'CS'
OTGW_CMD_CONTROL_HEATING = 'CH'
OTGW_CMD_VENT = 'VS'
OTGW_CMD_RST_CNT = 'RS'
OTGW_CMD_IGNORE_TRANS = 'IT'
OTGW_CMD_OVRD_HIGH = 'OH'
OTGW_CMD_OVRD_THRMST = 'FT'
OTGW_CMD_VREF = 'VR'

OTGW_MODE = 'otgw_mode'
OTGW_DHW_OVRD = 'otgw_dhw_ovrd'
OTGW_ABOUT = 'otgw_about'
OTGW_BUILD = 'otgw_build'
OTGW_CLOCKMHZ = 'otgw_clockmhz'
OTGW_LED_A = 'otgw_led_a'
OTGW_LED_B = 'otgw_led_b'
OTGW_LED_C = 'otgw_led_c'
OTGW_LED_D = 'otgw_led_d'
OTGW_LED_E = 'otgw_led_e'
OTGW_LED_F = 'otgw_led_f'
OTGW_GPIO_A = 'otgw_gpio_a'
OTGW_GPIO_B = 'otgw_gpio_b'
OTGW_GPIO_A_STATE = 'otgw_gpio_a_state'
OTGW_GPIO_B_STATE = 'otgw_gpio_b_state'
OTGW_SB_TEMP = 'otgw_setback_temp'
OTGW_SETP_OVRD_MODE = 'otgw_setpoint_ovrd_mode'
OTGW_SMART_PWR = 'otgw_smart_pwr'
OTGW_THRM_DETECT = 'otgw_thermostat_detect'
OTGW_IGNORE_TRANSITIONS = 'otgw_ignore_transitions'
OTGW_OVRD_HB = 'otgw_ovrd_high_byte'
OTGW_VREF = 'otgw_vref'

OTGW_SETP_OVRD_TEMPORARY = 'T'
OTGW_SETP_OVRD_PERMANENT = 'C'
OTGW_SETP_OVRD_DISABLED = 'N'
OTGW_MODE_MONITOR = 'M'
OTGW_MODE_GATEWAY = 'G'
OTGW_MODE_RESET = 'R'

OTGW_REPORT_ABOUT = 'A'
OTGW_REPORT_BUILDDATE = 'B'
OTGW_REPORT_CLOCKMHZ = 'C'
OTGW_REPORT_GPIO_FUNCS = 'G'
OTGW_REPORT_GPIO_STATES = 'I'
OTGW_REPORT_LED_FUNCS = 'L'
OTGW_REPORT_GW_MODE = 'M'
OTGW_REPORT_SETPOINT_OVRD = 'O'
OTGW_REPORT_SMART_PWR = 'P'
OTGW_REPORT_THERMOSTAT_DETECT = 'R'
OTGW_REPORT_SETBACK_TEMP = 'S'
OTGW_REPORT_TWEAKS = 'T'
OTGW_REPORT_VREF = 'V'
OTGW_REPORT_DHW_SETTING = 'W'

OTGW_REPORTS = {
    OTGW_REPORT_ABOUT: OTGW_ABOUT,
    OTGW_REPORT_BUILDDATE: OTGW_BUILD,
    OTGW_REPORT_CLOCKMHZ: OTGW_CLOCKMHZ,
    OTGW_REPORT_GPIO_FUNCS: [OTGW_GPIO_A, OTGW_GPIO_B],
    OTGW_REPORT_GPIO_STATES: [OTGW_GPIO_A_STATE, OTGW_GPIO_B_STATE],
    OTGW_REPORT_LED_FUNCS: [OTGW_LED_A, OTGW_LED_B, OTGW_LED_C,
                            OTGW_LED_D, OTGW_LED_E, OTGW_LED_F],
    OTGW_REPORT_GW_MODE: OTGW_MODE,
    OTGW_REPORT_SETPOINT_OVRD: OTGW_SETP_OVRD_MODE,
    OTGW_REPORT_SMART_PWR: OTGW_SMART_PWR,
    OTGW_REPORT_THERMOSTAT_DETECT: OTGW_THRM_DETECT,
    OTGW_REPORT_SETBACK_TEMP: OTGW_SB_TEMP,
    OTGW_REPORT_TWEAKS: [OTGW_IGNORE_TRANSITIONS, OTGW_OVRD_HB],
    OTGW_REPORT_VREF: OTGW_VREF,
    OTGW_REPORT_DHW_SETTING: OTGW_DHW_OVRD
}

NO_GOOD = 'NG'
SYNTAX_ERR = 'SE'
BAD_VALUE = 'BV'
OUT_OF_RANGE = 'OR'
NO_SPACE = 'NS'
NOT_FOUND = 'NF'
OVERRUN_ERR = 'OE'

OTGW_ERRS = {
    NO_GOOD: RuntimeError("No Good: The command code is unknown."),
    SYNTAX_ERR: SyntaxError("Syntax Error: The command contained an "
                            "unexpected character or was incomplete."),
    BAD_VALUE: ValueError("Bad Value: The command contained a data value that "
                          "is not allowed."),
    OUT_OF_RANGE: RuntimeError("Out of Range: A number was specified outside "
                               "of the allowed range."),
    NO_SPACE: RuntimeError("No Space: The alternative Data-ID could not be "
                           "added because the table is full."),
    NOT_FOUND: RuntimeError("Not Found: The specified alternative Data-ID "
                            "could not be removed because it does not exist "
                            "in the table."),
    OVERRUN_ERR: RuntimeError("Overrun Error: The processor was busy and "
                              "failed to process all received characters.")
}



# This file is part of pyotgw.
#
# pyotgw is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyotgw is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyotgw.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 Milan van Nugteren
#

import asyncio
import logging
import re
import struct
from asyncio.queues import QueueFull

from .vars import *

_LOGGER = logging.getLogger(__name__)


class protocol(asyncio.Protocol):
    """
    Implementation of the Opentherm Gateway protocol to be used with
    asyncio connections.
    """

    def connection_made(self, transport):
        """
        Gets called when a connection to the gateway is established.
        Initialise the protocol object.
        """
        self.transport = transport
        self.loop = transport.loop
        self._cmd_lock = asyncio.Lock(loop=self.loop)
        self._cmdq = asyncio.Queue(loop=self.loop)
        self._updateq = asyncio.Queue(loop=self.loop)
        self._readbuf = b''
        self._update_cb = None
        self._received_lines = 0
        self._watchdog_task = None
        self._report_task = None
        self.status = {}
        self.connected = True

    def connection_lost(self, exc):
        """
        Gets called when the connection to the gateway is lost.
        Tear down and clean up the protocol object.
        """
        _LOGGER.error("Disconnected: %s", exc)
        self.connected = False
        self.transport.close()
        for q in [self._cmdq, self._updateq]:
            while not q.empty():
                q.get()
        self._report_task.cancel()
        self.status = {}

    def data_received(self, data):
        """
        Gets called when new data is received on the serial interface.
        Perform line buffering and call line_received() with complete
        lines.
        """
        # DIY line buffering...
        newline = b'\r\n'
        eot = b'\x04'
        self._readbuf += data
        while newline in self._readbuf:
            line, _, self._readbuf = self._readbuf.partition(newline)
            if line:
                if eot in line:
                    # Discard everything before EOT
                    _, _, line = line.partition(eot)
                self.line_received(line.decode('ascii'))

    async def setup_watchdog(self, cb, timeout):
        """Trigger a reconnect after @timeout seconds of inactivity."""
        self._watchdog_timeout = timeout
        self._watchdog_cb = cb
        self._watchdog_task = self.loop.create_task(self._watchdog(timeout))

    async def cancel_watchdog(self):
        """Cancel the watchdog task and related variables."""
        if self._watchdog_task is not None:
            _LOGGER.debug("Cancelling Watchdog task.")
            self._watchdog_task.cancel()
            try:
                await self._watchdog_task
            except asyncio.CancelledError:
                self._watchdog_task = None

    async def _inform_watchdog(self):
        """Inform the watchdog of activity."""
        if self._watchdog_task is not None:
            self._watchdog_task.cancel()
            try:
                await self._watchdog_task
            except asyncio.CancelledError:
                self._watchdog_task = self.loop.create_task(self._watchdog(
                    self._watchdog_timeout))

    async def _watchdog(self, timeout):
        """Trigger and cancel the watchdog after timeout. Call callback."""
        await asyncio.sleep(timeout, loop=self.loop)
        _LOGGER.debug("Watchdog triggered!")
        await self.cancel_watchdog()
        await self._watchdog_cb()

    def line_received(self, line):
        """
        Gets called by data_received() when a complete line is
        received.
        Inspect the received line and process or queue accordingly.
        """
        self._received_lines += 1
        self.loop.create_task(self._inform_watchdog())
        pattern = r'^(T|B|R|A|E)([0-9A-F]{8})$'
        msg = re.match(pattern, line)
        if msg:
            mtype, mid, msb, lsb = self._dissect_msg(msg)
            if lsb is not None:
                self._process_msg(mtype, mid, msb, lsb)
        elif re.match(r'^[0-9A-F]{1,8}$', line) and self._received_lines == 1:
            # Partial message on fresh connection. Ignore.
            self._received_lines = 0
            pass
        else:
            try:
                self._cmdq.put_nowait(line)
            except QueueFull:
                _LOGGER.error('pyotgw: Queue full, discarded message: %s',
                              line)

    def _dissect_msg(self, match):
        """
        Split messages into bytes and return a tuple of bytes.
        """
        recvfrom = match.group(1)
        frame = bytes.fromhex(match.group(2))
        if recvfrom is 'E':
            print("pyotgw: Received erroneous message, ignoring:", frame)
            return (None, None, None, None)
        msgtype = self._get_msgtype(frame[0])
        if msgtype in (READ_ACK, WRITE_ACK, READ_DATA, WRITE_DATA):
            # Some info is best read from the READ/WRITE_DATA messages
            # as the boiler may not support the data ID.
            # Slice syntax is used to prevent implicit cast to int.
            data_id = frame[1:2]
            data_msb = frame[2:3]
            data_lsb = frame[3:4]
            # Ignore output to the thermostat ('A') except MSG_TROVRD,
            # MSG_TOUTSIDE and MSG_ROVRD as they may contain useful values.
            # Other messages cause issues when overriding values sent to the
            # boiler.
            if recvfrom is 'A' and data_id not in [MSG_TROVRD, MSG_TOUTSIDE,
                                                   MSG_ROVRD]:
                return (None, None, None, None)
            return (msgtype, data_id, data_msb, data_lsb)
        return (None, None, None, None)

    def _get_msgtype(self, byte):
        """
        Return the message type of Opentherm messages according to
        byte 1.
        """
        return (byte >> 4) & 0x7

    def _process_msg(self, msgtype, msgid, msb, lsb):
        """
        Process message and update status variables where necessary.
        Add status to queue if it was changed in the process.
        """
        if msgtype in (READ_DATA, WRITE_DATA):
            # Data sent from thermostat
            if msgid == MSG_STATUS:
                # Master sends status
                thermo_status = self._get_flag8(msb)
                self.status[DATA_MASTER_CH_ENABLED] = thermo_status[0]
                self.status[DATA_MASTER_DHW_ENABLED] = thermo_status[1]
                self.status[DATA_MASTER_COOLING_ENABLED] = thermo_status[2]
                self.status[DATA_MASTER_OTC_ENABLED] = thermo_status[3]
                self.status[DATA_MASTER_CH2_ENABLED] = thermo_status[4]
            elif msgid == MSG_MCONFIG:
                # Master sends ID
                self.status[DATA_MASTER_MEMBERID] = self._get_u8(lsb)
            elif msgid == MSG_TRSET:
                # Master changes room setpoint, support by the boiler
                # is not mandatory, but we want the data regardless
                self.status[DATA_ROOM_SETPOINT] = self._get_f8_8(msb, lsb)
            elif msgid == MSG_TRSET2:
                # Master changes room setpoint 2, support by the boiler
                # is not mandatory, but we want the data regardless
                self.status[DATA_ROOM_SETPOINT_2] = self._get_f8_8(msb, lsb)
            elif msgid == MSG_TROOM:
                # Master reports sensed room temperature
                self.status[DATA_ROOM_TEMP] = self._get_f8_8(msb, lsb)
            elif msgid == MSG_OTVERM:
                # Master reports OpenTherm version
                self.status[DATA_MASTER_OT_VERSION] = self._get_f8_8(msb, lsb)
            elif msgid == MSG_MVER:
                # Master reports product type and version
                self.status[DATA_MASTER_PRODUCT_TYPE] = self._get_u8(msb)
                self.status[DATA_MASTER_PRODUCT_VERSION] = self._get_u8(lsb)
        elif msgtype in (READ_ACK, WRITE_ACK):
            # Data sent from boiler
            if msgid == MSG_STATUS:
                # Slave reports status
                boiler_status = self._get_flag8(lsb)
                self.status[DATA_SLAVE_FAULT_IND] = boiler_status[0]
                self.status[DATA_SLAVE_CH_ACTIVE] = boiler_status[1]
                self.status[DATA_SLAVE_DHW_ACTIVE] = boiler_status[2]
                self.status[DATA_SLAVE_FLAME_ON] = boiler_status[3]
                self.status[DATA_SLAVE_COOLING_ACTIVE] = boiler_status[4]
                self.status[DATA_SLAVE_CH2_ACTIVE] = boiler_status[5]
                self.status[DATA_SLAVE_DIAG_IND] = boiler_status[6]
            elif msgid == MSG_TSET:
                # Slave confirms CH water setpoint
                self.status[DATA_CONTROL_SETPOINT] = self._get_f8_8(msb, lsb)
            elif msgid == MSG_SCONFIG:
                # Slave reports config and ID
                slave_status = self._get_flag8(msb)
                self.status[DATA_SLAVE_DHW_PRESENT] = slave_status[0]
                self.status[DATA_SLAVE_CONTROL_TYPE] = slave_status[1]
                self.status[DATA_SLAVE_COOLING_SUPPORTED] = slave_status[2]
                self.status[DATA_SLAVE_DHW_CONFIG] = slave_status[3]
                self.status[DATA_SLAVE_MASTER_LOW_OFF_PUMP] = slave_status[4]
                self.status[DATA_SLAVE_CH2_PRESENT] = slave_status[5]
                self.status[DATA_SLAVE_MEMBERID] = self._get_u8(lsb)
            elif msgid == MSG_COMMAND:
                # TODO: implement command notification system
                pass
            elif msgid == MSG_ASFFLAGS:
                # Slave reports fault flags
                fault_flags = self._get_flag8(msb)
                self.status[DATA_SLAVE_SERVICE_REQ] = fault_flags[0]
                self.status[DATA_SLAVE_REMOTE_RESET] = fault_flags[1]
                self.status[DATA_SLAVE_LOW_WATER_PRESS] = fault_flags[2]
                self.status[DATA_SLAVE_GAS_FAULT] = fault_flags[3]
                self.status[DATA_SLAVE_AIR_PRESS_FAULT] = fault_flags[4]
                self.status[DATA_SLAVE_WATER_OVERTEMP] = fault_flags[5]
                self.status[DATA_SLAVE_OEM_FAULT] = self._get_u8(lsb)
            elif msgid == MSG_RBPFLAGS:
                # Slave reports remote parameters
                transfer_flags = self._get_flag8(msb)
                rw_flags = self._get_flag8(lsb)
                self.status[DATA_REMOTE_TRANSFER_DHW] = transfer_flags[0]
                self.status[DATA_REMOTE_TRANSFER_MAX_CH] = transfer_flags[1]
                self.status[DATA_REMOTE_RW_DHW] = rw_flags[0]
                self.status[DATA_REMOTE_RW_MAX_CH] = rw_flags[1]
            elif msgid == MSG_COOLING:
                # Only report cooling control signal if slave acks it
                self.status[DATA_COOLING_CONTROL] = self._get_f8_8(msb, lsb)
            elif msgid == MSG_TSETC2:
                # Slave confirms CH2 water setpoint
                self.status[DATA_CONTROL_SETPOINT_2] = self._get_f8_8(msb, lsb)
            elif msgid == MSG_TROVRD:
                # OTGW (or downstream device) reports remote override
                ovrd_value = self._get_f8_8(msb, lsb)
                if ovrd_value > 0:
                    self.status[DATA_ROOM_SETPOINT_OVRD] = ovrd_value
                elif self.status.get(DATA_ROOM_SETPOINT_OVRD):
                    del self.status[DATA_ROOM_SETPOINT_OVRD]
            elif msgid == MSG_MAXRMOD:
                # Slave reports maximum modulation level
                self.status[DATA_SLAVE_MAX_RELATIVE_MOD] = self._get_f8_8(
                    msb, lsb)
            elif msgid == MSG_MAXCAPMINMOD:
                # Slave reports max capaxity and min modulation level
                self.status[DATA_SLAVE_MAX_CAPACITY] = self._get_u8(msb)
                self.status[DATA_SLAVE_MIN_MOD_LEVEL] = self._get_u8(lsb)
            elif msgid == MSG_RELMOD:
                # Slave reports relative modulation level
                self.status[DATA_REL_MOD_LEVEL] = self._get_f8_8(msb, lsb)
            elif msgid == MSG_CHPRESS:
                # Slave reports CH circuit pressure
                self.status[DATA_CH_WATER_PRESS] = self._get_f8_8(msb, lsb)
            elif msgid == MSG_DHWFLOW:
                # Slave reports DHW flow rate
                self.status[DATA_DHW_FLOW_RATE] = self._get_f8_8(msb, lsb)
            elif msgid == MSG_TBOILER:
                # Slave reports CH water temperature
                self.status[DATA_CH_WATER_TEMP] = self._get_f8_8(msb, lsb)
            elif msgid == MSG_TDHW:
                # Slave reports DHW temperature
                self.status[DATA_DHW_TEMP] = self._get_f8_8(msb, lsb)
            elif msgid == MSG_TOUTSIDE:
                # OTGW (or downstream device) reports outside temperature
                self.status[DATA_OUTSIDE_TEMP] = self._get_f8_8(msb, lsb)
            elif msgid == MSG_TRET:
                # Slave reports return water temperature
                self.status[DATA_RETURN_WATER_TEMP] = self._get_f8_8(msb, lsb)
            elif msgid == MSG_TSTOR:
                # Slave reports solar storage temperature
                self.status[DATA_SOLAR_STORAGE_TEMP] = self._get_f8_8(msb, lsb)
            elif msgid == MSG_TCOLL:
                # Slave reports solar collector temperature
                self.status[DATA_SOLAR_COLL_TEMP] = self._get_f8_8(msb, lsb)
            elif msgid == MSG_TFLOWCH2:
                # Slave reports CH2 water temperature
                self.status[DATA_CH_WATER_TEMP_2] = self._get_f8_8(msb, lsb)
            elif msgid == MSG_TDHW2:
                # Slave reports DHW2 temperature
                self.status[DATA_DHW_TEMP_2] = self._get_f8_8(msb, lsb)
            elif msgid == MSG_TEXHAUST:
                # Slave reports exhaust temperature
                self.status[DATA_EXHAUST_TEMP] = self._get_s16(msb, lsb)
            elif msgid == MSG_TDHWSETUL:
                # Slave reports min/max DHW setpoint
                self.status[DATA_SLAVE_DHW_MAX_SETP] = self._get_s8(msb)
                self.status[DATA_SLAVE_DHW_MIN_SETP] = self._get_s8(lsb)
            elif msgid == MSG_TCHSETUL:
                # Slave reports min/max CH setpoint
                self.status[DATA_SLAVE_CH_MAX_SETP] = self._get_s8(msb)
                self.status[DATA_SLAVE_CH_MIN_SETP] = self._get_s8(lsb)
            elif msgid == MSG_TDHWSET:
                # Slave reports or acks DHW setpoint
                self.status[DATA_DHW_SETPOINT] = self._get_f8_8(msb, lsb)
            elif msgid == MSG_MAXTSET:
                # Slave reports or acks max CH setpoint
                self.status[DATA_MAX_CH_SETPOINT] = self._get_f8_8(msb, lsb)
            elif msgid == MSG_ROVRD:
                # OTGW (or downstream device) reports remote override
                # behaviour
                rovrd_flags = self._get_flag8(lsb)
                self.status[DATA_ROVRD_MAN_PRIO] = rovrd_flags[0]
                self.status[DATA_ROVRD_AUTO_PRIO] = rovrd_flags[1]
            elif msgid == MSG_OEMDIAG:
                # Slave reports diagnostic info
                self.status[DATA_OEM_DIAG] = self._get_u16(msb, lsb)
            elif msgid == MSG_BURNSTARTS:
                # Slave reports burner starts
                self.status[DATA_TOTAL_BURNER_STARTS] = self._get_u16(msb, lsb)
            elif msgid == MSG_CHPUMPSTARTS:
                # Slave reports CH pump starts
                self.status[DATA_CH_PUMP_STARTS] = self._get_u16(msb, lsb)
            elif msgid == MSG_DHWPUMPSTARTS:
                # Slave reports DHW pump starts
                self.status[DATA_DHW_PUMP_STARTS] = self._get_u16(msb, lsb)
            elif msgid == MSG_DHWBURNSTARTS:
                # Slave reports DHW burner starts
                self.status[DATA_DHW_BURNER_STARTS] = self._get_u16(msb, lsb)
            elif msgid == MSG_BURNHRS:
                # Slave reports CH burner hours
                self.status[DATA_TOTAL_BURNER_HOURS] = self._get_u16(msb, lsb)
            elif msgid == MSG_CHPUMPHRS:
                # Slave reports CH pump hours
                self.status[DATA_CH_PUMP_HOURS] = self._get_u16(msb, lsb)
            elif msgid == MSG_DHWPUMPHRS:
                # Slave reports DHW pump hours
                self.status[DATA_DHW_PUMP_HOURS] = self._get_u16(msb, lsb)
            elif msgid == MSG_DHWBURNHRS:
                # Slave reports DHW burner hours
                self.status[DATA_DHW_BURNER_HOURS] = self._get_u16(msb, lsb)
            elif msgid == MSG_OTVERS:
                # Slave reports OpenTherm version
                self.status[DATA_SLAVE_OT_VERSION] = self._get_f8_8(msb, lsb)
            elif msgid == MSG_SVER:
                # Slave reports product type and version
                self.status[DATA_SLAVE_PRODUCT_TYPE] = self._get_u8(msb)
                self.status[DATA_SLAVE_PRODUCT_VERSION] = self._get_u8(lsb)
        self._updateq.put_nowait(self.status)

    def _get_flag8(self, byte):
        """
        Split a byte into a list of 8 bits (1/0).
        """
        ret = [0, 0, 0, 0, 0, 0, 0, 0]
        byte = byte[0]
        for i in range(0, 8):
            ret[i] = (byte & 1)
            byte = byte >> 1
        return ret

    def _get_u8(self, byte):
        """
        Convert a byte into an unsigned int.
        """
        return struct.unpack('>B', byte)[0]

    def _get_s8(self, byte):
        """
        Convert a byte into a signed int.
        """
        return struct.unpack('>b', byte)[0]

    def _get_f8_8(self, msb, lsb):
        """
        Convert 2 bytes into an OpenTherm f8_8 (float) value.
        """
        return float(self._get_s16(msb, lsb) / 256)

    def _get_u16(self, msb, lsb):
        """
        Convert 2 bytes into an unsigned int.
        """
        buf = struct.pack('>BB', self._get_u8(msb), self._get_u8(lsb))
        return int(struct.unpack('>H', buf)[0])

    def _get_s16(self, msb, lsb):
        """
        Convert 2 bytes into a signed int.
        """
        buf = struct.pack('>bB', self._get_s8(msb), self._get_u8(lsb))
        return int(struct.unpack('>h', buf)[0])

    async def _report(self):
        """
        Call _update_cb with the status dict as an argument whenever a status
        update occurs.

        This method is a coroutine
        """
        while True:
            oldstatus = dict(self.status)
            stat = await self._updateq.get()
            if self._update_cb is not None and oldstatus != stat:
                # Each client gets its own copy of the dict.
                self.loop.create_task(self._update_cb(dict(stat)))

    async def set_update_cb(self, cb):
        """Register the update callback."""
        if self._report_task is not None and not self._report_task.cancelled():
            self.loop.create_task(self._report_task.cancel())
        self._update_cb = cb
        if cb is not None:
            self._report_task = self.loop.create_task(self._report())

    async def issue_cmd(self, cmd, value, retry=3):
        """
        Issue a command, then await and return the return value.

        This method is a coroutine
        """
        async with self._cmd_lock:
            if not self.connected:
                _LOGGER.debug(
                    "Serial transport closed, not sending command %s", cmd)
                return
            while not self._cmdq.empty():
                _LOGGER.debug("Clearing leftover message from command queue:"
                              " %s", self._cmdq.get())
            _LOGGER.debug("Sending command: %s with value %s", cmd, value)
            self.transport.write(
                '{}={}\r\n'.format(cmd, value).encode('ascii'))
            expect = r'^{}:\s*([^$]+)$'.format(cmd)

            async def send_again(err):
                """Resend the command."""
                nonlocal retry
                _LOGGER.warning("Command %s failed with %s, retrying...", cmd,
                                err)
                retry -= 1
                self.transport.write(
                    '{}={}\r\n'.format(cmd, value).encode('ascii'))

            async def process(msg):
                """Process a possible response."""
                _LOGGER.debug("Got possible response for command %s: %s", cmd,
                              msg)
                if msg in OTGW_ERRS:
                    # Some errors appear by themselves on one line.
                    if retry == 0:
                        raise OTGW_ERRS[msg]
                    await send_again(msg)
                    return
                if cmd == OTGW_CMD_MODE and value == 'R':
                    # Device was reset, msg contains build info
                    while not re.match(
                            r'OpenTherm Gateway \d+\.\d+\.\d+', msg):
                        msg = await self._cmdq.get()
                    return True
                match = re.match(expect, msg)
                if match:
                    if match.group(1) in OTGW_ERRS:
                        # Some errors are considered a response.
                        if retry == 0:
                            raise OTGW_ERRS[match.group(1)]
                        await send_again(msg)
                        return
                    ret = match.group(1)
                    if cmd == OTGW_CMD_SUMMARY and ret == '1':
                        # Expects a second line
                        part2 = await self._cmdq.get()
                        ret = [ret, part2]
                    return ret
                if re.match(r'Error 0[1-4]', msg):
                    _LOGGER.warning("Received %s. If this happens during a "
                                    "reset of the gateway it can be safely "
                                    "ignored.", msg)
                    return
                _LOGGER.warning("Unknown message in command queue: %s", msg)
                await send_again(msg)

            while True:
                msg = await self._cmdq.get()
                ret = await process(msg)
                if ret is not None:
                    return ret
